from flask_babel import lazy_gettext as _l
from wtforms import BooleanField, PasswordField, SelectField, StringField, HiddenField
from wtforms.fields.html5 import EmailField, URLField
from wtforms.validators import InputRequired, Length

from CTFd.forms import BaseForm
from CTFd.forms.fields import SubmitField
from CTFd.models import Brackets, TeamFieldEntries, TeamFields
from CTFd.utils.countries import SELECT_COUNTRIES_LIST
from CTFd.utils.user import get_current_team

teamNameLength = 25


def build_team_bracket_field(form_cls, value=None):
    field = getattr(form_cls, "bracket_id", None)  # noqa B009
    if field:
        field.field_type = "select"
        field.process_data(value)
        return [field]
    else:
        return []


def attach_team_bracket_field(form_cls):
    brackets = Brackets.query.filter_by(type="teams").all()
    if brackets:
        choices = [("", "")] + [
            (bracket.id, f"{bracket.name} - {bracket.description}")
            for bracket in brackets
        ]
        select_field = SelectField(
            "Catégorie",
            description="Catégorie de compétition pour votre équipe",
            choices=choices,
            validators=[InputRequired()],
        )
        setattr(form_cls, "bracket_id", select_field)  # noqa B010


def build_custom_team_fields(
    form_cls,
    include_entries=False,
    fields_kwargs=None,
    field_entries_kwargs=None,
    blacklisted_items=(),
):
    if fields_kwargs is None:
        fields_kwargs = {}
    if field_entries_kwargs is None:
        field_entries_kwargs = {}

    fields = []
    new_fields = TeamFields.query.filter_by(**fields_kwargs).all()
    user_fields = {}

    # Only include preexisting values if asked
    if include_entries is True:
        for f in TeamFieldEntries.query.filter_by(**field_entries_kwargs).all():
            user_fields[f.field_id] = f.value

    for field in new_fields:
        if field.name.lower() in blacklisted_items:
            continue

        form_field = getattr(form_cls, f"fields[{field.id}]")

        # Add the field_type to the field so we know how to render it
        form_field.field_type = field.field_type

        # Only include preexisting values if asked
        if include_entries is True:
            initial = user_fields.get(field.id, "")
            form_field.data = initial
            if form_field.render_kw:
                form_field.render_kw["data-initial"] = initial
            else:
                form_field.render_kw = {"data-initial": initial}

        fields.append(form_field)
    return fields


def attach_custom_team_fields(form_cls, **kwargs):
    new_fields = TeamFields.query.filter_by(**kwargs).all()
    for field in new_fields:
        validators = []
        if field.required:
            validators.append(InputRequired())

        if field.field_type == "text":
            input_field = StringField(
                field.name, description=field.description, validators=validators
            )
        elif field.field_type == "boolean":
            input_field = BooleanField(
                field.name, description=field.description, validators=validators
            )

        setattr(form_cls, f"fields[{field.id}]", input_field)


class TeamJoinForm(BaseForm):
    name = StringField(_l("Nom de l'équipe"), validators=[InputRequired(), Length(max=teamNameLength)])
    password = PasswordField(_l("Mot de passe de l'équipe"), validators=[InputRequired()])
    submit = SubmitField(_l("Rejoindre"))


def TeamRegisterForm(*args, **kwargs):
    class _TeamRegisterForm(BaseForm):
        name = StringField(_l("Nom de l'équipe"), validators=[
            InputRequired(),
            Length(max=teamNameLength, message='Vous devez entrer moins de personnages')
        ], render_kw={"maxlength": teamNameLength})
        password = PasswordField(_l("Mot de passe de l'équipe"), validators=[InputRequired()])
        color = HiddenField(_l("Couleur de l'équipe"))
        submit = SubmitField(_l("Créer"))

        @property
        def extra(self):
            return build_custom_team_fields(
                self, include_entries=False, blacklisted_items=()
            ) + build_team_bracket_field(self)

    attach_custom_team_fields(_TeamRegisterForm)
    attach_team_bracket_field(_TeamRegisterForm)
    return _TeamRegisterForm(*args, **kwargs)


def TeamSettingsForm(*args, **kwargs):
    class _TeamSettingsForm(BaseForm):
        name = StringField(
            _l("Nom de l'équipe"),
            description=_l("Nom public de votre équipe montré à d'autres concurrents"),
        )
        password = PasswordField(
            _l("Nouveau mot de passe de l'équipe"), description=_l("Définissez un nouvel mot de passe de jointure de l'équipe")
        )
        color = HiddenField(
            _l("Couleur de l'équipe"),
            description=_l(
            "Couleur utilisée par l'équipe pour contrôler l'esthétique de l'équipe.")
        )
        confirm = PasswordField(
            _l("Confirmer le mot de passe actuel de l'équipe"),
            description=_l(
                "Fournissez le mot de passe actuel de l'équipe (ou votre mot de passe) pour mettre à jour le mot de passe de votre équipe"
            ),
        )
        affiliation = StringField(
            _l("Affiliation"),
            description=_l(
                "L'affiliation de votre équipe affichée publiquement aux autres concurrents"
            ),
        )
        website = URLField(
            _l("Site web"),
            description=_l("Le site Web de votre équipe affichée publiquement aux autres concurrents"),
        )
        country = SelectField(
            _l("Pays"),
            choices=SELECT_COUNTRIES_LIST,
            description=_l("Le pays de votre équipe affiché publiquement aux autres concurrents"),
        )
        submit = SubmitField(_l("Soumettre"))

        @property
        def extra(self):
            fields_kwargs = _TeamSettingsForm.get_field_kwargs()
            return build_custom_team_fields(
                self,
                include_entries=True,
                fields_kwargs=fields_kwargs,
                field_entries_kwargs={"team_id": self.obj.id},
            )

        def get_field_kwargs():
            team = get_current_team()
            field_kwargs = {"editable": True}
            if team.filled_all_required_fields is False:
                # Show all fields
                field_kwargs = {}
            return field_kwargs

        def __init__(self, *args, **kwargs):
            """
            Custom init to persist the obj parameter to the rest of the form
            """
            super().__init__(*args, **kwargs)
            obj = kwargs.get("obj")
            if obj:
                self.obj = obj

    field_kwargs = _TeamSettingsForm.get_field_kwargs()
    attach_custom_team_fields(_TeamSettingsForm, **field_kwargs)

    return _TeamSettingsForm(*args, **kwargs)


class TeamCaptainForm(BaseForm):
    # Choices are populated dynamically at form creation time
    captain_id = SelectField(
        _l("Capitaine d'équipe"), choices=[], validators=[InputRequired()]
    )
    submit = SubmitField("Soumettre")


class TeamSearchForm(BaseForm):
    field = SelectField(
        "Champ de recherche",
        choices=[
            ("name", "Nom"),
            ("id", "ID"),
            ("affiliation", "Affiliation"),
            ("website", "Site web"),
        ],
        default="name",
        validators=[InputRequired()],
    )
    q = StringField("Paramètre", validators=[InputRequired()])
    submit = SubmitField("Recherche")


class PublicTeamSearchForm(BaseForm):
    field = SelectField(
        _l("Champ de recherche"),
        choices=[
            ("name", _l("Nom")),
            ("affiliation", _l("Affiliation")),
            ("website", _l("Site web")),
        ],
        default="name",
        validators=[InputRequired()],
    )
    q = StringField(_l("Paramètre"), validators=[InputRequired()])
    submit = SubmitField(_l("Recherche"))


class TeamBaseForm(BaseForm):
    name = StringField(_l("Nom de l'équipe"), validators=[
            InputRequired(),
            Length(max=teamNameLength, message='Vous devez entrer moins de personnages')
        ], render_kw={"maxlength": teamNameLength})
    email = EmailField(_l("Courriel"))
    password = PasswordField(_l("Mot de passe"))
    color = HiddenField(_l("Couleur de l'équipe"))
    website = URLField(_l("Site web"))
    affiliation = StringField(_l("Affiliation"))
    country = SelectField(_l("Pays"), choices=SELECT_COUNTRIES_LIST)
    hidden = BooleanField(_l("Cachée"))
    banned = BooleanField(_l("Banni"))
    submit = SubmitField(_l("Soumettre"))


def TeamCreateForm(*args, **kwargs):
    class _TeamCreateForm(TeamBaseForm):
        pass

        @property
        def extra(self):
            return build_custom_team_fields(
                self, include_entries=False
            ) + build_team_bracket_field(self)

    attach_custom_team_fields(_TeamCreateForm)
    attach_team_bracket_field(_TeamCreateForm)

    return _TeamCreateForm(*args, **kwargs)


def TeamEditForm(*args, **kwargs):
    class _TeamEditForm(TeamBaseForm):
        pass

        @property
        def extra(self):
            return build_custom_team_fields(
                self,
                include_entries=True,
                fields_kwargs=None,
                field_entries_kwargs={"team_id": self.obj.id},
            ) + build_team_bracket_field(self, value=self.obj.bracket_id)

        def __init__(self, *args, **kwargs):
            """
            Custom init to persist the obj parameter to the rest of the form
            """
            super().__init__(*args, **kwargs)
            obj = kwargs.get("obj")
            if obj:
                self.obj = obj

    attach_custom_team_fields(_TeamEditForm)
    attach_team_bracket_field(_TeamEditForm)

    return _TeamEditForm(*args, **kwargs)


class TeamInviteForm(BaseForm):
    link = URLField(_l("Lien d'invitation"))


class TeamInviteJoinForm(BaseForm):
    submit = SubmitField(_l("Rejoindre"))
